@@ -2,7 +2,11 @@
 #this repository contains the full copyright notices and license terms.
 import gobject
 import gtk
-from tryton.rpc import RPCProxy
+import sys
+if sys.version_info < (2, 6):
+    import simplejson as json
+else:
+    import json
 import tryton.rpc as rpc
 import locale
 from interface import ParserView
@@ -13,6 +17,7 @@ import tryton.common as common
 from tryton.config import CONFIG
 from tryton.common.cellrendererbutton import CellRendererButton
 from tryton.common.cellrenderertoggle import CellRendererToggle
+from tryton.pyson import PYSONEncoder
 import os
 
 _ = gettext.gettext
@@ -20,48 +25,101 @@ _ = gettext.gettext
 
 class AdaptModelGroup(gtk.GenericTreeModel):
 
-    def __init__(self, model_group):
+    def __init__(self, group, children_field=None):
         super(AdaptModelGroup, self).__init__()
-        self.model_group = model_group
-        self.models = model_group.models
+        self.group = group
         self.last_sort = None
         self.sort_asc = True
         self.set_property('leak_references', False)
+        self.children_field = children_field
+        self.__removed = None # XXX dirty hack to allow update of has_child
 
-    def added(self, modellist, position):
-        if modellist is self.models:
-            model = self.models[position]
-            self.row_inserted(self.on_get_path(model),
-                    self.get_iter(self.on_get_path(model)))
+    def added(self, group, record):
+        if (group is self.group
+                and (record.group is self.group
+                    or record.group.child_name == self.children_field)):
+            path = self.on_get_path(record)
+            iter_ = self.get_iter(path)
+            self.row_inserted(path, iter_)
+            if record.children_group(self.children_field):
+                self.row_has_child_toggled(path, iter_)
+            if (record.parent and
+                    record.group is not self.group):
+                path = self.on_get_path(record.parent)
+                iter_ = self.get_iter(path)
+                self.row_has_child_toggled(path, iter_)
 
     def cancel(self):
         pass
 
-    def removed(self, lst, position):
-        self.row_deleted(position)
-        self.invalidate_iters()
+    def removed(self, group, record):
+        if (group is self.group
+                and (record.group is self.group
+                    or record.group.child_name == self.children_field)):
+            path = self.on_get_path(record)
+            self.row_deleted(path)
+            if (record.parent and
+                    record.group != self.group and
+                    len(record.children_group(self.children_field)) <= 1):
+                path = self.on_get_path(record.parent)
+                iter_ = self.get_iter(path)
+                self.__removed = record # XXX check for thread
+                self.row_has_child_toggled(path, iter_)
+                self.__removed = None
 
     def append(self, model):
-        self.model_group.model_add(model)
+        self.group.add(model)
 
     def prepend(self, model):
-        self.model_group.model_add(model, 0)
+        self.group.add(model, 0)
 
-    def remove(self, iter):
-        idx = self.get_path(iter)[0]
-        self.model_group.model_remove(self.models[idx])
+    def remove(self, iter_):
+        record = self.get_value(iter_, 0)
+        record.group.remove(record)
         self.invalidate_iters()
 
-    def move(self, path, position):
-        idx = path[0]
-        self.model_group.model_move(self.models[idx], position)
+    def __move(self, record, path, offset=0):
+        iter_ = self.get_iter(path)
+        record_pos = self.get_value(iter_, 0)
+        group = record_pos.group
+        pos = group.index(record_pos) + offset
+        if group is not record.group:
+            prev_group = record.group
+            record.group.remove(record, remove=True, force_remove=True)
+            # Don't remove record from previous group
+            # as the new parent will change the parent
+            # This prevents concurrency conflict
+            record.group.record_removed.remove(record)
+            group.add(record)
+            if not record.parent_name:
+                record.modified_fields.setdefault(prev_group.parent_name)
+                record.value[prev_group.parent_name] = False
+            else:
+                record.modified_fields.setdefault(record.parent_name)
+        group.move(record, pos)
+
+    def move_before(self, record, path):
+        self.__move(record, path)
+
+    def move_after(self, record, path):
+        self.__move(record, path, 1)
+
+    def move_into(self, record, path):
+        iter_ = self.get_iter(path)
+        parent = self.get_value(iter_, 0)
+        group = parent.children_group(self.children_field, check_load=True)
+        if group is not record.group:
+            record.group.remove(record, remove=True, force_remove=True)
+            group.add(record)
+            record.modified_fields.setdefault(record.parent_name or 'id')
+        group.move(record, 0)
 
     def sort(self, ids):
         ids2pos = {}
         pos = 0
         new_order = []
-        for model in self.models:
-            ids2pos[model.id] = pos
+        for record in self.group:
+            ids2pos[record.id] = pos
             new_order.append(pos)
             pos += 1
         pos = 0
@@ -71,97 +129,138 @@ class AdaptModelGroup(gtk.GenericTreeModel):
                 if old_pos != pos:
                     new_order[old_pos] = pos
                 pos += 1
-            except:
+            except Exception:
                 continue
-        self.models.sort(lambda x, y: \
+        self.group.sort(lambda x, y: \
                 cmp(new_order[ids2pos[x.id]], new_order[ids2pos[y.id]]))
         prev = None
-        for model in self.models:
+        for record in self.group:
             if prev:
-                prev.next[id(self.models)] = model
-            prev = model
+                prev.next[id(self.group)] = record
+            prev = record
         if prev:
-            prev.next[id(self.models)] = None
+            prev.next[id(self.group)] = None
         self.rows_reordered(None, None, new_order)
 
     def __len__(self):
-        return len(self.models)
+        return len(self.group)
 
     def on_get_flags(self):
-        return gtk.TREE_MODEL_LIST_ONLY
+        if not self.children_field:
+            return gtk.TREE_MODEL_LIST_ONLY
+        return 0
 
     def on_get_n_columns(self):
+        # XXX
         return 1
 
     def on_get_column_type(self, index):
+        # XXX
         return gobject.TYPE_PYOBJECT
 
-    def on_get_path(self, iter):
-        if iter in self.models:
-            return self.models.index(iter)
+    def on_get_path(self, iter_):
+        if isinstance(iter_, tuple):
+            return tuple(x[0] for x in iter_)
         else:
-            return 0
+            path = []
+            i = iter_
+            while i:
+                path.append(i.group.index(i))
+                if i.group is self.group:
+                    break
+                i = i.parent
+            path.reverse()
+            return tuple(path)
+
+    def on_get_tree_path(self, iter):
+        return self.on_get_path(iter)
 
     def on_get_iter(self, path):
-        if isinstance(path, tuple):
-            path = path[0]
-        if self.models:
-            if path < len(self.models):
-                return self.models[path]
-            else:
+        group = self.group
+        record = None
+        for i in path:
+            if group is None or i >= len(group):
                 return None
-        else:
+            record = group[i]
+            if not self.children_field:
+                break
+            group = record.children_group(self.children_field, check_load=True)
+        return record
+
+    def on_get_value(self, record, column):
+        return record
+
+    def on_iter_next(self, record):
+        if record is None:
             return None
+        return record.next.get(id(record.group))
 
-    def on_get_value(self, node, column):
-        assert column == 0
-        return node
+    def on_iter_has_child(self, record):
+        if record is None or not self.children_field:
+            return False
+        children = record.children_group(self.children_field)
+        length = len(children)
+        if self.__removed and self.__removed in children:
+            length -= 1
+        return bool(length)
 
-    def on_iter_next(self, node):
-        try:
-            return node.next[id(self.models)]
-        except IndexError:
+    def on_iter_children(self, record):
+        if record is None:
             return None
-
-    def on_iter_has_child(self, node):
-        return False
-
-    def on_iter_children(self, node):
+        if self.children_field:
+            children = record.children_group(self.children_field)
+            if children:
+                return children[0]
         return None
 
-    def on_iter_n_children(self, node):
-        if node is None:
-            return len(self.models)
-        return 0
+    def on_iter_n_children(self, record):
+        if record is None or not self.children_field:
+            return len(self.group)
+        return len(record.children_group(self.children_field))
 
-    def on_iter_nth_child(self, node, nth):
-        if node is None and self.models:
-            return self.on_get_iter(0)
+    def on_iter_nth_child(self, record, nth):
+        if record is None or not self.children_field:
+            if nth < len(self.group):
+                return self.group[nth]
+            return None
+        if nth < len(record.children_group(self.children_field)):
+            return record.children_group(self.children_field)[nth]
         return None
 
-    def on_iter_parent(self, node):
-        return None
+    def on_iter_parent(self, record):
+        if record is None:
+            return None
+        return record.parent
 
 
 class ViewList(ParserView):
 
     def __init__(self, window, screen, widget, children=None, buttons=None,
-            toolbar=None, notebooks=None, cursor_widget=None):
+            toolbar=None, notebooks=None, cursor_widget=None,
+            children_field=None):
         super(ViewList, self).__init__(window, screen, widget, children,
-                buttons, toolbar, notebooks, cursor_widget)
+                buttons, toolbar, notebooks, cursor_widget, children_field)
         self.store = None
         self.view_type = 'tree'
-        self.model_add_new = True
-        self.widget = gtk.VBox()
-        self.widget_tree = widget
+
+        vbox = gtk.VBox()
         scroll = gtk.ScrolledWindow()
-        scroll.add(self.widget_tree)
+        scroll.add(self.widget)
         scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
+        scroll.set_placement(gtk.CORNER_TOP_LEFT)
         viewport = gtk.Viewport()
         viewport.set_shadow_type(gtk.SHADOW_ETCHED_IN)
         viewport.add(scroll)
-        self.widget.pack_start(viewport, expand=True, fill=True)
+        self.widget_tree = self.widget
+
+        if isinstance(self.screen.window, gtk.Dialog):
+            width, height = self.widget_tree.size_request()
+            vbox.set_size_request(width or -1, height or -1)
+        vbox.pack_start(viewport, expand=True, fill=True)
+
         self.widget_tree.screen = screen
+
+        self.widget = vbox
         self.reload = False
         self.children = children
 
@@ -192,10 +291,14 @@ class ViewList(ParserView):
                     continue
 
                 for tool in toolbar[icontype]:
-                    iconstock = {
-                        'print': 'tryton-print',
-                        'action': 'tryton-executable',
-                    }.get(icontype)
+                    if not tool['icon.rec_name']:
+                        iconstock = {
+                            'print': 'tryton-print',
+                            'action': 'tryton-executable',
+                        }.get(icontype)
+                    else:
+                        iconstock = tool['icon.rec_name']
+                    common.ICONFACTORY.register_icon(iconstock)
 
                     if hasattr(gtk, 'MenuToolButton') and icontype == 'print':
                         tbutton = gtk.MenuToolButton(iconstock)
@@ -239,26 +342,40 @@ class ViewList(ParserView):
         selection.set_mode(gtk.SELECTION_MULTIPLE)
         selection.connect('changed', self.__select_changed)
 
+        dnd = False
+        if self.children_field:
+            children_field = self.widget_tree.cells.get(self.children_field)
+            if children_field:
+                parent_name = children_field.attrs.get('relation_field')
+                dnd = parent_name in self.widget_tree.cells
+        elif self.widget_tree.sequence:
+            dnd = True
         # Disable DnD on mac until it is fully supported
-        if self.widget_tree.sequence \
-                and not (os.name == 'mac' \
-                    or (hasattr(os, 'uname') and os.uname()[0] == 'Darwin')):
-            self.widget_tree.enable_model_drag_source(gtk.gdk.BUTTON1_MASK,
-                    [('MY_TREE_MODEL_ROW', gtk.TARGET_SAME_WIDGET, 0),],
-                    gtk.gdk.ACTION_MOVE)
+        if sys.platform == 'darwin':
+            dnd = False
+        if screen.readonly:
+            dnd = False
+        if dnd:
             self.widget_tree.drag_source_set(gtk.gdk.BUTTON1_MASK | gtk.gdk.BUTTON3_MASK,
                     [('MY_TREE_MODEL_ROW', gtk.TARGET_SAME_WIDGET, 0),],
                     gtk.gdk.ACTION_MOVE)
-            self.widget_tree.enable_model_drag_dest(
+            self.widget_tree.drag_dest_set(gtk.DEST_DEFAULT_ALL,
                     [('MY_TREE_MODEL_ROW', gtk.TARGET_SAME_WIDGET, 0),],
                     gtk.gdk.ACTION_MOVE)
 
+            self.widget_tree.connect('drag-begin', self.drag_begin)
+            self.widget_tree.connect('drag-motion', self.drag_motion)
             self.widget_tree.connect('drag-drop', self.drag_drop)
             self.widget_tree.connect("drag-data-get", self.drag_data_get)
             self.widget_tree.connect('drag-data-received', self.drag_data_received)
             self.widget_tree.connect('drag-data-delete', self.drag_data_delete)
 
-        self.widget_tree.connect('key_press_event', self.on_keypres)
+        self.widget_tree.connect('key_press_event', self.on_keypress)
+        self.widget_tree.connect('test-expand-row', self.test_expand_row)
+        self.widget_tree.set_expander_column(self.widget_tree.get_column(0))
+
+    def get_fields(self):
+        return [col.name for col in self.widget_tree.get_columns() if col.name]
 
     def _sig_clicked(self, widget, action, atype):
         return self._action(action, atype)
@@ -271,27 +388,81 @@ class ViewList(ParserView):
             message(_('No record selected!'), self.window)
             return False
         email = {}
-        if action.get('email'):
-            email = self.screen.current_model.expr_eval(action['email'])
+        if action.get('pyson_email'):
+            email = self.screen.current_record.expr_eval(action['pyson_email'])
             if not email:
                 email = {}
         email['subject'] = action['name'].replace('_', '')
         act['email'] = email
         data = {
-            'model': self.screen.name,
+            'model': self.screen.model_name,
             'id': obj_id,
             'ids': obj_ids,
         }
         value = Action._exec_action(act, self.window, data, {})
         if self.screen:
-            self.screen.reload(writen=True)
+            self.screen.reload(written=True)
         return value
 
 
-    def on_keypres(self, widget, event):
+    def on_keypress(self, widget, event):
         if event.keyval == gtk.keysyms.c and event.state & gtk.gdk.CONTROL_MASK:
             self.on_copy()
             return False
+        if event.keyval in (gtk.keysyms.Down, gtk.keysyms.Up):
+            path, column = widget.get_cursor()
+            if not path:
+                return False
+            model = widget.get_model()
+            if event.keyval == gtk.keysyms.Down:
+                test = True
+                for i in xrange(len(path)):
+                    iter_ = model.get_iter(path[0:i+1])
+                    if model.iter_next(iter_):
+                        test = False
+                if test:
+                    iter_ = model.get_iter(path)
+                    if (model.iter_has_child(iter_)
+                            and widget.row_expanded(path)):
+                        test = False
+                return test
+            elif event.keyval == gtk.keysyms.Up:
+                if path == (0,):
+                    return True
+        if event.keyval in (gtk.keysyms.Left, gtk.keysyms.Right):
+            selection = widget.get_selection()
+            model, paths = selection.get_selected_rows()
+            if event.keyval == gtk.keysyms.Left:
+                if len(paths) == 1:
+                    path, = paths
+                    if not widget.row_expanded(path):
+                        path = path[:-1]
+                        if path:
+                            selection.select_path(path)
+                            widget.collapse_row(path)
+                for path in paths:
+                    widget.collapse_row(path)
+            elif event.keyval == gtk.keysyms.Right:
+                for path in paths:
+                    widget.expand_row(path, False)
+
+    def test_expand_row(self, widget, iter_, path):
+        model = widget.get_model()
+        iter_ = model.iter_children(iter_)
+        if not iter_:
+            return False
+        fields = [col.name for col in self.widget_tree.get_columns()
+                if col.name]
+        while iter_:
+            record = model.get_value(iter_, 0)
+            if not record.get_loaded(fields):
+                try:
+                    for field in fields:
+                        record.__getitem__(field, True)
+                except Exception, exception:
+                    return True
+            iter_ = model.iter_next(iter_)
+        return False
 
     def on_copy(self):
         clipboard = self.widget_tree.get_clipboard(gtk.gdk.SELECTION_CLIPBOARD)
@@ -307,8 +478,7 @@ class ViewList(ParserView):
         # Don't use set_with_data on mac see:
         # http://bugzilla.gnome.org/show_bug.cgi?id=508601
         if selection.count_selected_rows() < 100 \
-                or os.name == 'mac' \
-                or (hasattr(os, 'uname') and os.uname()[0] == 'Darwin'):
+                or sys.platform == 'darwin':
             data = []
             selection.selected_foreach(self.copy_foreach, data)
             clipboard.set_text('\n'.join(data))
@@ -317,13 +487,13 @@ class ViewList(ParserView):
                     self.copy_clear_func, selection)
 
     def copy_foreach(self, treemodel, path,iter, data):
-        model = treemodel.get_value(iter, 0)
+        record = treemodel.get_value(iter, 0)
         values = []
         for col in self.widget_tree.get_columns():
             if not col.get_visible() or not col.name:
                 continue
             cell = self.widget_tree.cells[col.name]
-            values.append('"' + str(cell.get_textual_value(model)) + '"')
+            values.append('"' + str(cell.get_textual_value(record)) + '"')
         data.append('\t'.join(values))
         return
 
@@ -338,6 +508,22 @@ class ViewList(ParserView):
         del selection
         return
 
+    def drag_begin(self, treeview, context):
+        return True
+
+    def drag_motion(self, treeview, context, x, y, time):
+        try:
+            treeview.set_drag_dest_row(*treeview.get_dest_row_at_pos(x, y))
+        except TypeError:
+            treeview.set_drag_dest_row(len(treeview.get_model()) - 1,
+                gtk.TREE_VIEW_DROP_AFTER)
+        if context.get_source_widget() == treeview:
+            kind = gtk.gdk.ACTION_MOVE
+        else:
+            kind = gtk.gdk.ACTION_COPY
+        context.drag_status(kind, time)
+        return True
+
     def drag_drop(self, treeview, context, x, y, time):
         treeview.emit_stop_by_name('drag-drop')
         treeview.drag_get_data(context, context.targets[-1], time)
@@ -346,8 +532,9 @@ class ViewList(ParserView):
     def drag_data_get(self, treeview, context, selection, target_id,
             etime):
         treeview.emit_stop_by_name('drag-data-get')
-        def _func_sel_get(store, path, iter, data):
-            data.append(path)
+        def _func_sel_get(store, path, iter_, data):
+            value = store.get_value(iter_, 0)
+            data.append(json.dumps(value.get_path(store.group)))
         data = []
         treeselection = treeview.get_selection()
         treeselection.selected_foreach(_func_sel_get, data)
@@ -355,31 +542,55 @@ class ViewList(ParserView):
             return
         data = str(data[0])
         selection.set(selection.target, 8, data)
+        return True
 
     def drag_data_received(self, treeview, context, x, y, selection,
             info, etime):
         treeview.emit_stop_by_name('drag-data-received')
         if treeview.sequence:
-            for model in self.screen.models.models:
-                if model[treeview.sequence].get_state_attrs(
-                        model).get('readonly', False):
+            field = self.screen.group.fields[treeview.sequence]
+            for record in self.screen.group:
+                if field.get_state_attrs(
+                        record).get('readonly', False):
                     return
         if not selection.data:
             return
-        model = treeview.get_model()
-        data = common.safe_eval(selection.data)
+        store = treeview.get_model()
+        try:
+            data = json.loads(selection.data)
+        except ValueError:
+            return
+        record = store.group.get_by_path(data)
+        record_path = store.on_get_path(record)
         drop_info = treeview.get_dest_row_at_pos(x, y)
+        def check_recursion(from_, to):
+            if not from_ or not to:
+                return True
+            if from_ == to:
+                return False
+            length = min(len(from_), len(to))
+            if len(from_) < len(to) and from_[:length] == to[:length]:
+                return False
+            return True
         if drop_info:
             path, position = drop_info
-            idx = path[0]
-            if position in (gtk.TREE_VIEW_DROP_BEFORE,
-                    gtk.TREE_VIEW_DROP_INTO_OR_BEFORE):
-                model.move(data, idx)
-            else:
-                model.move(data, idx + 1)
+            check_path = path
+            if position in (gtk.TREE_VIEW_DROP_BEFORE, gtk.TREE_VIEW_DROP_AFTER):
+                check_path = path[:-1]
+            if not check_recursion(record_path, check_path):
+                return
+            if position == gtk.TREE_VIEW_DROP_BEFORE:
+                store.move_before(record, path)
+            elif position == gtk.TREE_VIEW_DROP_AFTER:
+                store.move_after(record, path)
+            elif self.children_field:
+                store.move_into(record, path)
+        else:
+            store.move_after(record, (len(store) - 1,))
         context.drop_finish(False, etime)
         if treeview.sequence:
-            self.screen.models.set_sequence(field=treeview.sequence)
+            record.group.set_sequence(field=treeview.sequence)
+        return True
 
     def drag_data_delete(self, treeview, context):
         treeview.emit_stop_by_name('drag-data-delete')
@@ -394,18 +605,20 @@ class ViewList(ParserView):
                 model = selection.get_selected_rows()[0]
             if (not path) or not path[0]:
                 return False
-            model = model.models[path[0][0]]
+            record = model.group[path[0][0]]
 
             if hasattr(path[1], '_type') and path[1]._type == 'many2one':
-                value = model[path[1].name].get(model)
-                ir_action_keyword = RPCProxy('ir.action.keyword')
+                value = record[path[1].name].get(record)
+                args = ('model', 'ir.action.keyword', 'get_keyword',
+                        'form_relate', (self.screen.group.fields[
+                            path[1].name].attrs['relation'], 0), rpc.CONTEXT)
                 try:
-                    relates = ir_action_keyword.get_keyword('form_relate',
-                            (self.screen.fields[path[1].name]['relation'], 0),
-                            rpc.CONTEXT)
+                    relates = rpc.execute(*args)
                 except Exception, exception:
-                    common.process_exception(exception, self.window)
-                    return False
+                    relates = common.process_exception(exception, self.window,
+                            *args)
+                    if not relates:
+                        return False
                 menu_entries = []
                 menu_entries.append((None, None, None))
                 menu_entries.append((_('Actions'),
@@ -444,38 +657,43 @@ class ViewList(ParserView):
                     self.window)
             return False
         from tryton.gui.window.view_form.screen import Screen
-        screen = Screen(self.screen.fields[path[1].name]['relation'],
-                self.window)
+        screen = Screen(self.screen.group.fields[
+            path[1].name].attrs['relation'], self.window)
         screen.load([value])
-        act['domain'] = str(screen.current_model.expr_eval(act.get('domain', []),
-                check_load=False))
-        act['context'] = str(screen.current_model.expr_eval(act.get('context', {}),
-            check_load=False))
-        data['model'] = self.screen.name
+        encoder = PYSONEncoder()
+        act['domain'] = encoder.encode(screen.current_record.expr_eval(
+            act.get('domain', []), check_load=False))
+        act['context'] = encoder.encode(screen.current_record.expr_eval(
+            act.get('context', {}), check_load=False))
+        data['model'] = self.screen.model_name
         data['id'] = value
         data['ids'] = [value]
         return Action._exec_action(act, self.window, data, context)
 
     def click_and_action(self, atype, value, path):
         return Action.exec_keyword(atype, self.window, {
-            'model': self.screen.fields[path[1].name]['relation'],
+            'model': self.screen.group.fields[
+                path[1].name].attrs['relation'],
             'id': value or False,
             'ids': [value],
             }, alwaysask=True)
 
-    def signal_record_changed(self, signal, *args):
-        if self.store:
-            if signal == 'record-added':
-                self.store.added(*args)
-            elif signal == 'record-removed':
-                self.store.removed(*args)
-        self.update_children()
+    def group_list_changed(self, group, signal):
+        if self.store is not None:
+            if signal[0] == 'record-added':
+                self.store.added(group, signal[1])
+            elif signal[0] == 'record-removed':
+                self.store.removed(group, signal[1])
+            elif signal[0] == 'group-cleared':
+                self.store = None
+                self.widget_tree.set_model(self.store)
+        self.display()
 
     def cancel(self):
         pass
 
     def __str__(self):
-        return 'ViewList (%s)' % self.screen.resource
+        return 'ViewList (%d)' % id(self)
 
     def __getitem__(self, name):
         return None
@@ -498,41 +716,54 @@ class ViewList(ParserView):
 
             if fields and any(fields.itervalues()):
                 try:
-                    rpc.execute('model', 'ir.ui.view_tree_width',
-                            'set_width', self.screen.name, fields, rpc.CONTEXT)
-                except:
+                    rpc.execute('model', 'ir.ui.view_tree_width', 'set_width',
+                            self.screen.model_name, fields, rpc.CONTEXT)
+                except Exception:
                     pass
         self.widget_tree.destroy()
-        del self.screen
-        del self.widget_tree
-        del self.widget
+        self.screen = None
+        self.widget_tree = None
+        self.widget = None
 
     def __sig_switch(self, treeview, path, column):
         if column._type == 'button':
             return
-        self.screen.row_activate()
+        if not self.screen.row_activate():
+            if treeview.row_expanded(path):
+                treeview.collapse_row(path)
+            else:
+                treeview.expand_row(path, False)
 
     def __select_changed(self, tree_sel):
-        previous_model = self.screen.current_model
+        previous_record = self.screen.current_record
+        if previous_record and previous_record not in previous_record.group:
+            previous_record = None
 
         if tree_sel.get_mode() == gtk.SELECTION_SINGLE:
-            model, iter = tree_sel.get_selected()
-            if iter:
-                path = model.get_path(iter)[0]
-                self.screen.current_model = model.models[path]
+            model, iter_ = tree_sel.get_selected()
+            if model and iter_:
+                record = model.get_value(iter_, 0)
+                self.screen.current_record = record
+            else:
+                self.screen.current_record = None
 
         elif tree_sel.get_mode() == gtk.SELECTION_MULTIPLE:
             model, paths = tree_sel.get_selected_rows()
-            if paths:
-                self.screen.current_model = model.models[paths[0][0]]
+            if model and paths:
+                iter_ = model.get_iter(paths[0])
+                record = model.get_value(iter_, 0)
+                self.screen.current_record = record
+            else:
+                self.screen.current_record = None
 
         if hasattr(self.widget_tree, 'editable') \
                 and self.widget_tree.editable \
-                and self.screen.tree_saves \
-                and previous_model != self.screen.current_model:
-            if previous_model and \
-                    not (previous_model.validate() and previous_model.save()):
-                self.screen.current_model = previous_model
+                and not self.screen.parent \
+                and previous_record != self.screen.current_record:
+            if previous_record and \
+                    not (previous_record.validate(self.get_fields())
+                            and previous_record.save()):
+                self.screen.current_record = previous_record
                 self.set_cursor()
                 return True
         self.update_children()
@@ -551,12 +782,13 @@ class ViewList(ParserView):
     def display(self):
         if self.reload \
                 or (not self.widget_tree.get_model()) \
-                    or self.screen.models != \
-                        self.widget_tree.get_model().model_group:
-            self.store = AdaptModelGroup(self.screen.models)
+                    or self.screen.group != \
+                        self.widget_tree.get_model().group:
+            self.store = AdaptModelGroup(self.screen.group,
+                    self.children_field)
             self.widget_tree.set_model(self.store)
         self.reload = False
-        if not self.screen.current_model:
+        if not self.screen.current_record:
             # Should find a simpler solution to do something like
             #self.widget.set_cursor(None,None,False)
             if self.store:
@@ -568,12 +800,12 @@ class ViewList(ParserView):
         self.update_children()
 
     def set_state(self):
-        model = self.screen.current_model
-        if model:
-            for field in model.mgroup.fields:
-                modelfield = model.mgroup.mfields.get(field, None)
-                if modelfield:
-                    modelfield.state_set(model)
+        record = self.screen.current_record
+        if record:
+            for field in record.group.fields:
+                field = record.group.fields.get(field, None)
+                if field:
+                    field.state_set(record)
 
     def update_children(self):
         ids = self.sel_ids_get()
@@ -581,23 +813,23 @@ class ViewList(ParserView):
             value = 0.0
             value_selected = 0.0
             loaded = True
-            for model in self.screen.models.models:
-                if not model.loaded:
+            for record in self.screen.group:
+                if not record.loaded:
                     loaded = False
                     break
-                if model.id in ids or not ids:
+                if record.id in ids or not ids:
                     if not value_selected:
-                        value_selected = model.fields_get()[self.children[child][0]]\
-                                .get(model, check_load=False)
+                        value_selected = record.fields_get()[self.children[child][0]]\
+                                .get(record, check_load=False)
                     else:
-                        value_selected += model.fields_get()[self.children[child][0]]\
-                                .get(model, check_load=False)
+                        value_selected += record.fields_get()[self.children[child][0]]\
+                                .get(record, check_load=False)
                 if not value:
-                    value = model.fields_get()[self.children[child][0]]\
-                            .get(model, check_load=False)
+                    value = record.fields_get()[self.children[child][0]]\
+                            .get(record, check_load=False)
                 else:
-                    value += model.fields_get()[self.children[child][0]]\
-                            .get(model, check_load=False)
+                    value += record.fields_get()[self.children[child][0]]\
+                            .get(record, check_load=False)
 
             if loaded:
                 label_str = locale.format('%.' + str(self.children[child][3]) + 'f',
@@ -607,15 +839,14 @@ class ViewList(ParserView):
                         value, True)
             else:
                 label_str = '-'
-            if self.children[child][4]:
-                self.children[child][2].set_markup('<b>%s</b>' % label_str)
-            else:
-                self.children[child][2].set_markup(label_str)
+            self.children[child][2].set_text(label_str)
 
     def set_cursor(self, new=False, reset_view=True):
         self.widget_tree.grab_focus()
-        if self.screen.current_model:
-            path = self.store.on_get_path(self.screen.current_model)
+        if self.screen.current_record:
+            path = self.store.on_get_path(self.screen.current_record)
+            if self.store.get_flags() & gtk.TREE_MODEL_LIST_ONLY:
+                path = (path[0],)
             focus_column = None
             for column in self.widget_tree.get_columns():
                 renderers = column.get_cell_renderers()
@@ -625,33 +856,36 @@ class ViewList(ParserView):
                 if isinstance(renderer, CellRendererToggle):
                     editable = renderer.get_property('activatable')
                 elif isinstance(renderer,
-                        (gtk.CellRendererProgress, CellRendererButton)):
+                        (gtk.CellRendererProgress, CellRendererButton,
+                            gtk.CellRendererPixbuf)):
                     editable = False
                 else:
                     editable = renderer.get_property('editable')
                 if column.get_visible() and editable:
                     focus_column = column
                     break
-            self.widget_tree.set_cursor(path, focus_column, new)
+            if path[:-1]:
+                self.widget_tree.expand_to_path(path[:-1])
             self.widget_tree.scroll_to_cell(path, focus_column, use_align=False)
+            self.widget_tree.set_cursor(path, focus_column, new)
 
     def sel_ids_get(self):
         def _func_sel_get(store, path, iter, ids):
-            model = store.on_get_iter(path)
-            if model and model.id > 0:
-                ids.append(model.id)
+            record = store.on_get_iter(path)
+            if record and record.id > 0:
+                ids.append(record.id)
         ids = []
         sel = self.widget_tree.get_selection()
         sel.selected_foreach(_func_sel_get, ids)
         return ids
 
-    def sel_models_get(self):
-        def _func_sel_get(store, path, iter, models):
-            models.append(store.on_get_iter(path))
-        models = []
+    def selected_records(self):
+        def _func_sel_get(store, path, iter, records):
+            records.append(store.on_get_iter(path))
+        records = []
         sel = self.widget_tree.get_selection()
-        sel.selected_foreach(_func_sel_get, models)
-        return models
+        sel.selected_foreach(_func_sel_get, records)
+        return records
 
     def unset_editable(self):
         self.widget_tree.editable = False
@@ -660,7 +894,8 @@ class ViewList(ParserView):
                 if isinstance(renderer, CellRendererToggle):
                     renderer.set_property('activatable', False)
                 elif isinstance(renderer,
-                        (gtk.CellRendererProgress, CellRendererButton)):
+                        (gtk.CellRendererProgress, CellRendererButton,
+                            gtk.CellRendererPixbuf)):
                     pass
                 else:
                     renderer.set_property('editable', False)
