@@ -14,14 +14,10 @@ import os
 _ = gettext.gettext
 
 
-
 class WinExport(object):
     "Window export"
 
-    def __init__(self, model, ids, fields, preload=None, parent=None,
-            context=None):
-        if preload is None:
-            preload = []
+    def __init__(self, model, ids, parent, context=None):
         self.dialog = gtk.Dialog(
                 title= _("Export to CSV"),
                 parent=parent,
@@ -169,6 +165,7 @@ class WinExport(object):
 
         self.view1 = gtk.TreeView()
         self.view1.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
+        self.view1.connect('row-expanded', self.on_row_expanded)
         scrolledwindow_all_fields.add(self.view1)
         self.view2 = gtk.TreeView()
         self.view2.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
@@ -177,49 +174,21 @@ class WinExport(object):
         self.view2.set_headers_visible(False)
 
         cell = gtk.CellRendererText()
-        column = gtk.TreeViewColumn('Field name', cell, text=0)
+        column = gtk.TreeViewColumn('Field name', cell, text=0,
+                background=2)
         self.view1.append_column(column)
 
         cell = gtk.CellRendererText()
         column = gtk.TreeViewColumn('Field name', cell, text=0)
         self.view2.append_column(column)
 
-        self.model1 = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
+        self.model1 = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_STRING,
+                gobject.TYPE_STRING)
         self.model2 = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
 
-        for i in preload:
-            self.model2.set(self.model2.append(), 0, i[1], 1, i[0])
-
         self.fields = {}
 
-        def model_populate(fields, prefix_node='', prefix=None,
-                prefix_value='', level=2):
-            fields_order = fields.keys()
-            fields_order.sort(lambda x, y: -cmp(fields[x].get('string', ''),
-                fields[y].get('string', '')))
-            for field in fields_order:
-                self.fields_data[prefix_node+field] = fields[field]
-                if prefix_node:
-                    self.fields_data[prefix_node + field]['string'] = \
-                            '%s%s' % (prefix_value,
-                                    self.fields_data[prefix_node + \
-                                            field]['string'])
-                st_name = fields[field]['string'] or field
-                node = self.model1.insert(prefix, 0,
-                        [st_name, prefix_node+field])
-                self.fields[prefix_node+field] = (st_name,
-                        fields[field].get('relation', False))
-                if fields[field].get('relation', False) and level>0:
-                    try:
-                        fields2 = rpc.execute('model',
-                                fields[field]['relation'], 'fields_get', None,
-                                rpc.CONTEXT)
-                    except Exception, exception:
-                        common.process_exception(exception, self.parent)
-                        continue
-                    model_populate(fields2, prefix_node+field+'/', node,
-                            st_name+'/', level-1)
-        model_populate(fields)
+        self.model_populate(self._get_fields(model))
 
         self.view1.set_model(self.model1)
         self.view2.set_model(self.model2)
@@ -232,10 +201,8 @@ class WinExport(object):
 
         # Creating the predefined export view
         self.pref_export = gtk.TreeView()
-        self.pref_export.append_column(gtk.TreeViewColumn(_('Export name'),
+        self.pref_export.append_column(gtk.TreeViewColumn(_('Name'),
             gtk.CellRendererText(), text=2))
-        self.pref_export.append_column(gtk.TreeViewColumn(_('Exported fields'),
-            gtk.CellRendererText(), text=3))
         scrolledwindow_exports.add(self.pref_export)
 
         self.pref_export.connect("row-activated", self.sel_predef)
@@ -244,17 +211,46 @@ class WinExport(object):
         self.predef_model = gtk.ListStore(
                 gobject.TYPE_INT,
                 gobject.TYPE_PYOBJECT,
-                gobject.TYPE_STRING,
                 gobject.TYPE_STRING)
         self.fill_predefwin()
         self.pref_export.show_all()
 
-    def sig_sel_all(self, widget=None):
-        self.model2.clear()
-        for field, relation in self.fields.keys():
-            if not relation:
-                self.model2.set(self.model2.append(), 0, self.fields[field],
-                        1, field)
+    def model_populate(self, fields, parent_node=None, prefix_field='',
+            prefix_name=''):
+        fields_order = fields.keys()
+        fields_order.sort(lambda x, y: -cmp(fields[x].get('string', ''),
+            fields[y].get('string', '')))
+        for field in fields_order:
+            self.fields_data[prefix_field + field] = fields[field]
+            name = fields[field]['string'] or field
+            if prefix_field:
+                self.fields_data[prefix_field + field]['string'] = '%s%s' % \
+                    (prefix_name, self.fields_data[prefix_field +
+                        field]['string'])
+            node = self.model1.insert(parent_node, 0, [name, prefix_field +
+                field, (fields[field].get('required', False) and
+                    common.COLORS['required']) or 'white'])
+            self.fields[prefix_field + field] = (name,
+                    fields[field].get('relation'))
+            if fields[field].get('relation'):
+                self.model1.insert(node, 0, [None, '', 'white'])
+
+    def _get_fields(self, model):
+        args = ('model', model, 'fields_get', None, rpc.CONTEXT)
+        try:
+            return rpc.execute(*args)
+        except Exception, exception:
+            return common.process_exception(exception, self.parent, *args)
+
+    def on_row_expanded(self, treeview, iter, path):
+        child = self.model1.iter_children(iter)
+        if self.model1.get_value(child, 0) is None:
+            prefix_field = self.model1.get_value(iter, 1)
+            _, model = self.fields[prefix_field]
+            name = self.fields_data[prefix_field]['string']
+            self.model_populate(self._get_fields(model), iter, prefix_field +
+                    '/', name + '/')
+            self.model1.remove(child)
 
     def sig_sel(self, widget=None):
         sel = self.view1.get_selection()
@@ -265,8 +261,8 @@ class WinExport(object):
         if relation:
             return
         num = self.model2.append()
-        self.model2.set(num, 0, store.get_value(iter, 0), 1,
-                store.get_value(iter, 1))
+        name = self.fields_data[store.get_value(iter, 1)]['string']
+        self.model2.set(num, 0, name, 1, store.get_value(iter, 1))
 
     def sig_unsel(self, widget=None):
         store, paths = self.view2.get_selection().get_selected_rows()
@@ -278,50 +274,68 @@ class WinExport(object):
         self.model2.clear()
 
     def fill_predefwin(self):
-        ir_export = rpc.RPCProxy('ir.export')
-        ir_export_line = rpc.RPCProxy('ir.export.line')
+        args = ('model', 'ir.export', 'search',
+                [('resource', '=', self.model)], 0, None, None, rpc.CONTEXT)
         try:
-            export_ids = ir_export.search([('resource', '=', self.model)])
+            export_ids = rpc.execute(*args)
         except Exception, exception:
-            common.process_exception(exception, self.parent)
-            return
-        for export in ir_export.read(export_ids):
-            try:
-                fields = ir_export_line.read(export['export_fields'])
-            except Exception, exception:
-                common.process_exception(exception, self.parent)
-                continue
+            export_ids = common.process_exception(exception, self.parent,
+                    *args)
+            if not export_ids:
+                return
+        args = ('model', 'ir.export', 'read', export_ids, None, rpc.CONTEXT)
+        try:
+            exports = rpc.execute(*args)
+        except Exception, exception:
+            exports = common.process_exception(exception, self.parent, *args)
+            if not exports:
+                return
+        args = ('model', 'ir.export.line', 'read',
+                sum((x['export_fields'] for x in exports), []), None,
+                rpc.CONTEXT)
+        try:
+            lines = rpc.execute(*args)
+        except Exception, exception:
+            lines = common.process_exception(exception, self.parent, *args)
+            if not lines:
+                return
+        id2lines = {}
+        for line in lines:
+            id2lines.setdefault(line['export'], []).append(line)
+        for export in exports:
             self.predef_model.append((
                 export['id'],
-                [f['name'] for f in fields],
-                export['name'],
-                ', '.join(self.fields_data[f['name']]['string'] \
-                        for f in fields),
-                ))
+                [x['name'] for x in id2lines.get(export['id'], [])],
+                export['name']))
         self.pref_export.set_model(self.predef_model)
 
     def add_predef(self, widget):
-        name = common.ask('What is the name of this export?', self.parent)
+        name = common.ask(_('What is the name of this export?'), self.parent)
         if not name:
             return
-        ir_export = rpc.RPCProxy('ir.export')
         iter = self.model2.get_iter_root()
         fields = []
         while iter:
             field_name = self.model2.get_value(iter, 1)
             fields.append(field_name)
             iter = self.model2.iter_next(iter)
+        args = ('model', 'ir.export', 'create', {
+            'name': name,
+            'resource': self.model,
+            'export_fields': [('create', {
+                'name': x,
+                }) for x in fields],
+            }, rpc.CONTEXT)
         try:
-            new_id = ir_export.create({'name' : name, 'resource' : self.model,
-                'export_fields' : [('create', {'name' : f}) for f in fields]})
+            new_id = rpc.execute(*args)
         except Exception, exception:
-            common.process_exception(exception, self.dialog)
-            return
+            new_ids = common.process_exception(exception, self.dialog, *args)
+            if not new_id:
+                return
         self.predef_model.append((
             new_id,
             fields,
-            name,
-            ','.join(self.fields_data[f]['string'] for f in fields)))
+            name))
         self.pref_export.set_model(self.predef_model)
 
     def remove_predef(self, widget):
@@ -331,13 +345,13 @@ class WinExport(object):
         (model, i) = sel
         if not i:
             return None
-        ir_export = rpc.RPCProxy('ir.export')
         export_id = model.get_value(i, 0)
+        args = ('model', 'ir.export', 'delete', export_id, rpc.CONTEXT)
         try:
-            ir_export.delete(export_id)
+            rpc.execute(*args)
         except Exception, exception:
-            common.process_exception(exception, self.dialog)
-            return
+            if not common.process_exception(exception, self.dialog, *args):
+                return
         for i in range(len(self.predef_model)):
             if self.predef_model[i][0] == export_id:
                 del self.predef_model[i]
@@ -347,6 +361,23 @@ class WinExport(object):
     def sel_predef(self, widget, path, column):
         self.model2.clear()
         for field in self.predef_model[path[0]][1]:
+            if field not in self.fields_data:
+                iter = self.model1.get_iter_first()
+                prefix = ''
+                for parent in field.split('/')[:-1]:
+                    while iter:
+                        if self.model1.get_value(iter, 1) == \
+                                (prefix + parent):
+                            self.on_row_expanded(self.view1, iter,
+                                    self.model1.get_path(iter))
+                            iter = self.model1.iter_children(iter)
+                            prefix = parent + '/'
+                            break
+                        else:
+                            iter = self.model1.iter_next(iter)
+
+            if field not in self.fields_data:
+                continue
             self.model2.append((self.fields_data[field]['string'], field))
 
     def run(self):
@@ -369,8 +400,8 @@ class WinExport(object):
                 fileno, fname = tempfile.mkstemp('.csv', 'tryton_')
                 self.export_csv(fname, fields2, result,
                         self.wid_write_field_names.get_active(), popup=False)
-                common.file_open(fname, 'csv', self.parent)
                 os.close(fileno)
+                common.file_open(fname, 'csv', self.parent)
             else:
                 fname = common.file_selection(_('Save As...'),
                         parent=self.parent,
@@ -386,7 +417,7 @@ class WinExport(object):
 
     def export_csv(self, fname, fields, result, write_title=False, popup=True):
         try:
-            file_p = file(fname, 'wb+')
+            file_p = open(fname, 'wb+')
             writer = csv.writer(file_p)
             if write_title:
                 writer.writerow(fields)
